@@ -245,9 +245,10 @@ update_account_list (CallsAccountOverview *self)
       CallsAccountRow *account_row = calls_account_row_new (provider, account);
 
       g_signal_handlers_disconnect_by_data (account, self);
-      g_signal_connect (account, "message",
-                        G_CALLBACK (on_account_message),
-                        self);
+      g_signal_connect_object (account, "message",
+                               G_CALLBACK (on_account_message),
+                               self,
+                               G_CONNECT_AFTER);
 
       gtk_list_box_insert (GTK_LIST_BOX (self->overview),
                            GTK_WIDGET (account_row),
@@ -272,12 +273,16 @@ on_providers_changed (CallsAccountOverview *self)
 
     if (CALLS_IS_ACCOUNT_PROVIDER (provider)) {
       self->providers = g_list_append (self->providers, provider);
-      g_signal_connect_swapped (calls_provider_get_origins (provider),
-                                "items-changed",
-                                G_CALLBACK (update_account_list),
-                                self);
-      g_signal_connect_swapped (provider, "widget-edit-done",
-                                G_CALLBACK (gtk_widget_hide), self->account_window);
+      g_signal_connect_object (calls_provider_get_origins (provider),
+                               "items-changed",
+                               G_CALLBACK (update_account_list),
+                               self,
+                               G_CONNECT_SWAPPED);
+      g_signal_connect_object (provider,
+                               "widget-edit-done",
+                               G_CALLBACK (gtk_widget_hide),
+                               self->account_window,
+                               G_CONNECT_SWAPPED);
     }
   }
 
